#include <stdio.h>      /* for printf() and fprintf() */
#include <sys/socket.h> /* for socket(), connect(), send(), and recv() */
#include <arpa/inet.h>  /* for sockaddr_in and inet_addr() */
#include <stdlib.h>     /* for atoi() and exit() */
#include <string.h>     /* for memset() */
#include <unistd.h>     /* for close() */

#define RCVBUFSIZE 32   /* Size of receive buffer */

void DieWithError(char *errorMessage);  /* Error handling function */

int main(int argc, char *argv[])
{
    int sock;                        /* Socket descriptor */
    struct sockaddr_in ServAddr;     /* Server address */
    unsigned short ServPort;         /* Server port */
    char *servIP;                    /* Server IP address */
    char *String;                    /* String to send to Server */
    char Buffer[RCVBUFSIZE];         /* Buffer for string */
    unsigned int StringLen;          /* Length of string to echo */
    int bytesRcvd, totalBytesRcvd;   /* Bytes read in single recv() 
                                        and total bytes read */

    if (argc != 4)    /* Correct number of input arguments? */
    {
       fprintf(stderr, "Usage: %s <Server IP> <Server Port> <Echo Word>\n",
               argv[0]);
       exit(1);
    }

    servIP   = argv[1];       /* First  arg: Server IP address */
    ServPort = atoi(argv[2]); /* Second arg: Server port */
    String   = argv[3];       /* Third  arg: string to echo */

    /* [1] SOCKET: Create a reliable, stream socket using TCP */

    if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");

    /* Construct the local representation of Server address structure */

    memset(&ServAddr, 0, sizeof(ServAddr));       /* Clear structure */
    ServAddr.sin_family      = AF_INET;           /* Address family: Internet */
    ServAddr.sin_addr.s_addr = inet_addr(servIP); /* Server IP address */
    ServAddr.sin_port        = htons(ServPort);   /* Server port */

    /* [2] CONNECT: Establish the connection to Server */

    if (connect(sock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        DieWithError("connect() failed");

    StringLen = strlen(String);          /* Determine input string lenght */

    /* [3.a] SEND: Send input string to Server */

    if (send(sock, String, StringLen, 0) != StringLen)
        DieWithError("send() sent a different number of bytes than expected");

    /* [3.b] RECV: Receive the same string back from Server */

    totalBytesRcvd = 0;                  /* Setup to print the echoed string */
    printf("Received echo message: "); 
    while (totalBytesRcvd < StringLen)
    {
        /* Receive up to the buffer size (minus 1 to leave space for
           null terminator) bytes from Server */

        if ((bytesRcvd = recv(sock, Buffer, RCVBUFSIZE - 1, 0)) <= 0)
            DieWithError("recv() failed or connection closed prematurely");
        totalBytesRcvd += bytesRcvd;   /* Keep tally of total bytes */
        Buffer[bytesRcvd] = '\0';      /* Terminate the string! */
        printf(Buffer);                /* Print the echo buffer */
    }

    printf("\n");    /* Print a final linefeed */

    /* [4] CLOSE: Close socket */

    close(sock);

    exit(0);
}
