#include <stdio.h>      /* for printf() and fprintf() */
#include <sys/socket.h> /* for socket(), bind(), and connect() */
#include <arpa/inet.h>  /* for sockaddr_in and inet_ntoa() */
#include <stdlib.h>     /* for atoi() and exit() */
#include <string.h>     /* for memset() */
#include <unistd.h>     /* for close() */

#define MAXPENDING 5    /* Maximum outstanding connection requests */

void DieWithError(char *errorMessage);  /* Error handling function */
void HandleTCPClient(int clntSocket);   /* TCP client handling function */

int main(int argc, char *argv[])
{
    int servSock;                    /* Server socket descriptor */
    int clntSock;                    /* Client socket descriptor */
    struct sockaddr_in ServAddr;     /* Server communication structure */
    struct sockaddr_in ClntAddr;     /* Local rep. of Client comm. structure */
    unsigned short ServPort;         /* Server port */
    unsigned int clntLen;            /* Size of Client data structure */

    if (argc != 2)     /* Correct numer of input arguments? */
    {
        fprintf(stderr, "Usage:  %s <Server Port>\n", argv[0]);
        exit(1);
    }

    ServPort = atoi(argv[1]);  /* Server port */

    /* [1] SOCKET: Create Server socket for incoming connections */

    if ((servSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");
      
    /* Construct local (Server) communication structure */
    memset(&ServAddr, 0, sizeof(ServAddr));       /* Clear Server structure */
    ServAddr.sin_family = AF_INET;                /* Address family: Internet */
    ServAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    ServAddr.sin_port = htons(ServPort);          /* Server port */

    /* [2] BIND: Bind Server port */

    if (bind(servSock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        DieWithError("bind() failed");

    /* [3] LISTEN: Make the Server socket listen for incoming connections */

    if (listen(servSock, MAXPENDING) < 0)
        DieWithError("listen() failed");

    /* Run forever */
    for (;;)
    {
        /* Set the size of the local representation of the Client structure */
        clntLen = sizeof(ClntAddr);

        /* [4] ACCEPT: Await request for connecting Server and Client sockets */

        if ((clntSock = accept(servSock, 
			       (struct sockaddr *) &ClntAddr, &clntLen)) < 0)
            DieWithError("accept() failed");

        /* Now clntSock is connected to a Client! */

        printf("Handling client located @ address %s : ", 
	       inet_ntoa(ClntAddr.sin_addr));

        HandleTCPClient(clntSock);
    }
}
